#include <string.h>
#include "lmp.h"

#ifdef DEBUG
#include <stdio.h>
#endif

/* global error report variable */
unsigned short lmp_proc_error = LMP_NO_ERROR;

/* return LMP version string */
char * lmp_ver () {
  static char version[] = LMP_VERSION;
#ifdef DEBUG
  fprintf (stderr, "(DBG) ver: RETURN %s\n", version);
#endif

  return version;
}

/* Evaluate the top level expressions: braces or numbers
 *
 * Create new instance to evaluate the braced-in expression
 * and return the result.
 * If no braces are available, we will just return the next
 * number.
 * returns the read or computed value
 */
lmp_value_type lmp_eval_top () { 

  lmp_value_type res;

  if (lmp_following_equals ("(")) {
#ifdef DEBUG
    fprintf (stderr, "(DBG) eval_top: Evaluating braces.\n");
#endif
    lmp_inc_token ();
    res = lmp_eval ();
    if (!lmp_following_equals (")"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_top: Missing closing brace.\n");
#endif
      lmp_error ("\")\" expected at argument %d\n", 0);
#ifdef DEBUG
    }
#endif
    lmp_inc_token ();

#ifdef DEBUG
    fprintf (stderr, "(DBG) eval_top: RETURN %Lf\n", res);
#endif
    return res; /* return inner value */
  }

  res = lmp_next_number ();

#ifdef DEBUG
  fprintf (stderr, "(DBG) eval_top: RETURN %Lf\n", res);
#endif

  return res; /* return next number */
}

/* Evaluate level 8 expressions
 *
 * This function parses and evaluates unary operations
 * like sin, cos, lg, sqrt etc.
 * It calls the top level evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_value_type lmp_eval_8 () {
  lmp_value_type operand;
  lmp_operator op;

  for (;;) {
    if (lmp_following_equals ("sin"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"sin\" catched.\n");
#endif
      op = LMP_SINE;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("cos"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"cos\" catched.\n");
#endif
      op = LMP_COSINE;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("tan"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"tan\" catched.\n");
#endif
      op = LMP_TANGENT;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("cot"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"cot\" catched.\n");
#endif
      op = LMP_COTANGENT;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("arcsin"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"arcsin\" catched.\n");
#endif
      op = LMP_ARC_SINE;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("arccos"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"arccos\" catched.\n");
#endif
      op = LMP_ARC_COSINE;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("arctan"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"arctan\" catched.\n");
#endif
      op = LMP_ARC_TANGENT;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("ln"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"ln\" catched.\n");
#endif
      op = LMP_LOGARITHM;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("lg"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"lg\" catched.\n");
#endif
      op = LMP_LOG_10;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("exp"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"exp\" catched.\n");
#endif
      op = LMP_EXP;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("sqrt"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_8: \"sqrt\" catched.\n");
#endif
      op = LMP_SQRT;
#ifdef DEBUG
    }
#endif
    else
#ifdef DEBUG
    {
      lmp_value_type result = lmp_eval_top ();
      fprintf (stderr, "(DBG) eval_8: RETURN %Lf\n", result);
      return result;
    }
#else
      return lmp_eval_top ();
#endif

    lmp_inc_token ();
    operand = lmp_unary_proc (op, lmp_eval_top ());
    if (lmp_proc_error != 0)
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) lmp_eval_8: Got lmp_proc_error.\n");
#endif
      lmp_stop (lmp_last_token ());
#ifdef DEBUG
    }
    fprintf (stderr, "(DBG) lmp_eval_8: RETURN %Lf\n", operand);
#endif
    return operand;
  }
}

/* Evaluate level 7 expressions
 *
 * This function parses and evaluates binary prefix
 * operation log. It calls the 8 level evaluation routine
 * for the arguments.
 * returns the computed value.
 */
lmp_value_type lmp_eval_7 () {
  lmp_value_type first, second;
  lmp_operator op;

  for (;;) {
    if (lmp_following_equals ("log"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_7: \"log\" catched.\n");
#endif
      op = LMP_LOG;
#ifdef DEBUG
    }
#endif
    else
#ifdef DEBUG
    {
      lmp_value_type result = lmp_eval_8 ();
      fprintf (stderr, "(DBG) eval_7: RETURN %Lf\n", result);
      return result;
    }
#else
    return lmp_eval_8 ();
#endif

    lmp_inc_token ();
    first = lmp_binary_proc (op, lmp_eval_8 (), lmp_eval_8 ());
    if (lmp_proc_error != 0)
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) lmp_eval_8: Got lmp_proc_error.\n");
#endif
      lmp_stop (lmp_last_token ());
#ifdef DEBUG
    }
    fprintf (stderr, "(DBG) lmp_eval_8: RETURN %Lf\n", first);
#endif
    return first;
  }

  return first;
}

/* Evaluate level 6 expressions
 *
 * This function parses and evaluates the binary operation
 * "^" (power).
 * It calls the level 7 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_value_type lmp_eval_6 () {
  lmp_value_type left, right;
  lmp_operator op;

  left = lmp_eval_7 ();

#ifdef DEBUG
  fprintf (stderr, "(DBG) eval_6: left -> %Lf\n", left);
#endif

  for (;;) {
    if (lmp_following_equals ("^"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_6: \"^\" catched.\n");
#endif
      op = LMP_POWER;
#ifdef DEBUG
    }
#endif
    else
      return left;

    lmp_inc_token ();
    right = lmp_eval_7 ();
    left = lmp_binary_proc (op, left, right);
    if (lmp_proc_error != 0)
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) lmp_eval_6: Got lmp_proc_error.\n");
#endif
      lmp_stop (lmp_last_token ());
#ifdef DEBUG
    }
#endif
  }

#ifdef DEBUG
  fprintf (stderr, "(DBG) lmp_eval_6: RETURN %Lf\n", left);
#endif
  return left;
}

/* Evaluate level 5 expressions
 *
 * This function parses and evaluates the binary operations
 * mod (modulo) and div (integer div).
 * It calls the level 6 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_value_type lmp_eval_5 () {
  lmp_value_type left, right;
  lmp_operator op;

  left = lmp_eval_6 ();

#ifdef DEBUG
  fprintf (stderr, "(DBG) eval_5: left -> %Lf\n", left);
#endif

  for (;;) {
    if (lmp_following_equals ("mod"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_5: \"mod\" catched.\n");
#endif
      op = LMP_MODULO;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("div"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_5: \"div\" catched.\n");
#endif
      op = LMP_DIV_NO_REM;
#ifdef DEBUG
    }
#endif
    else
      return left;

    lmp_inc_token ();
    right = lmp_eval_6 ();
    left = lmp_binary_proc (op, left, right);
    if (lmp_proc_error != 0)
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) lmp_eval_5: Got lmp_proc_error.\n");
#endif
      lmp_stop (lmp_last_token ());
#ifdef DEBUG
    }
#endif
  }

#ifdef DEBUG
  fprintf (stderr, "(DBG) lmp_eval_5: RETURN %Lf\n", left);
#endif
  return left;
}

/* Evaluate level 4 expressions
 *
 * This function parses and evaluates the binary operations
 * * and /.
 * It calls the level 5 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_value_type lmp_eval_4 () {
  lmp_value_type left, right;
  lmp_operator op;

  left = lmp_eval_5 ();

#ifdef DEBUG
  fprintf (stderr, "(DBG) eval_4: left -> %Lf\n", left);
#endif

  for (;;) {
    if (lmp_following_equals ("*"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_4: \"*\" catched.\n");
#endif
      op = LMP_MULTIPLICATION;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("/"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_4: \"/\" catched.\n");
#endif
      op = LMP_DIVISION;
#ifdef DEBUG
    }
#endif
    else
      return left;

    lmp_inc_token ();
    right = lmp_eval_5 ();
    left = lmp_binary_proc (op, left, right);
    if (lmp_proc_error != 0)
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) lmp_eval_4: Got lmp_proc_error.\n");
#endif
      lmp_stop (lmp_last_token ());
#ifdef DEBUG
    }
#endif
  }

#ifdef DEBUG
  fprintf (stderr, "(DBG) lmp_eval_4: RETURN %Lf\n", left);
#endif
  return left;
}

/* Evaluate level 3 expressions
 *
 * This function parses and evaluates the binary operations
 * + and -.
 * It calls the level 4 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_value_type lmp_eval_3 () {
  lmp_value_type left, right;
  lmp_operator op;

  left = lmp_eval_4 ();

#ifdef DEBUG
  fprintf (stderr, "(DBG) eval_3: left -> %Lf\n", left);
#endif

  for (;;) {
    if (lmp_following_equals ("+"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_3: \"+\" catched.\n");
#endif
      op = LMP_ADDITION;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("-"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_3: \"-\" catched.\n");
#endif
      op = LMP_SUBSTRACTION;
#ifdef DEBUG
    }
#endif
    else
      return left;

    lmp_inc_token ();
    right = lmp_eval_3 ();
    left = lmp_binary_proc (op, left, right);
    if (lmp_proc_error != 0)
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) lmp_eval_3: Got lmp_proc_error.\n");
#endif
      lmp_stop (lmp_last_token ());
#ifdef DEBUG
    }
#endif
  }

#ifdef DEBUG
  fprintf (stderr, "(DBG) lmp_eval_3: RETURN %Lf\n", left);
#endif
  return left;
}

/* Evaluate level 2 expressions
 *
 * This function parses and evaluates comparisons
 * like ==, < etc.
 * It calls the level 3 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_value_type lmp_eval_2 () {
  lmp_value_type left, right;
  lmp_operator op;

  left = lmp_eval_3 ();

#ifdef DEBUG
  fprintf (stderr, "(DBG) eval_2: left -> %Lf\n", left);
#endif

  for (;;) {
    if (lmp_following_equals ("=="))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_2: \"==\" catched.\n");
#endif
      op = LMP_EQUALS;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("!="))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_2: \"!=\" catched.\n");
#endif
      op = LMP_UNEQUALS;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals (">="))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_2: \">=\" catched.\n");
#endif
      op = LMP_GREATER_EQUAL;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("<="))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_2: \"<=\" catched.\n");
#endif
      op = LMP_LOWER_EQUAL;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals (">"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_2: \">\" catched.\n");
#endif
      op = LMP_GREATER;
#ifdef DEBUG
    }
#endif
    else if (lmp_following_equals ("<"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_2: \"<\" catched.\n");
#endif
      op = LMP_LOWER;
#ifdef DEBUG
    }
#endif
    else
      return left;

    lmp_inc_token ();
    right = lmp_eval_3 ();
    left = lmp_binary_proc (op, left, right);
    if (lmp_proc_error != 0)
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) lmp_eval_2: Got lmp_proc_error.\n");
#endif
      lmp_stop (lmp_last_token ());
#ifdef DEBUG
    }
#endif
  }

#ifdef DEBUG
  fprintf (stderr, "(DBG) lmp_eval_2: RETURN %Lf\n", left);
#endif
  return left;
}

/* Evaluate level 1 expressions
 *
 * This function parses and evaluates logical operation
 * and.
 * It calls the level 2 evaluation routine for the
 * argument.
 * returns the computed value.
 */
lmp_value_type lmp_eval_1 () {
  lmp_value_type left, right;
  lmp_operator op;

  left = lmp_eval_2 ();

#ifdef DEBUG
  fprintf (stderr, "(DBG) eval_1: left -> %Lf \n", left);
#endif

  for (;;) {
    if (lmp_following_equals ("and"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval_1: \"and\" catched.\n");
#endif
      op = LMP_OUTER_AND;
#ifdef DEBUG
    }
#endif
    else
      return left;

    lmp_inc_token ();
    right = lmp_eval_2 ();
    left = lmp_binary_proc (op, left, right);
    if (lmp_proc_error != 0)
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) lmp_eval_1: Got lmp_proc_error.\n");
#endif
      lmp_stop (lmp_last_token ());
#ifdef DEBUG
    }
#endif
  }

#ifdef DEBUG
  fprintf (stderr, "(DBG) lmp_eval_1: RETURN %Lf\n", left);
#endif
  return left;
}

/* Evaluate level 0 expressions
 *
 * This function parses and evaluates logical operation
 * "or" and calls the level 2 evaluation routine.
 * returns the computed value.
 */
lmp_value_type lmp_eval () {
  lmp_value_type left, right;
  lmp_operator op;

  left = lmp_eval_1 ();

#ifdef DEBUG
  fprintf (stderr, "(DBG) eval: left -> %Lf \n", left);
#endif

  for (;;) {
    if (lmp_following_equals ("or"))
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval: \"or\" catched.\n");
#endif
      op = LMP_OUTER_OR;
#ifdef DEBUG
    }
#endif
    else
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) eval: leave with %Lf\n", left);
#endif
      return left;
#ifdef DEBUG
    }
#endif

    lmp_inc_token ();
    right = lmp_eval_1 ();
    left = lmp_binary_proc (op, left, right);
    if (lmp_proc_error != 0)
#ifdef DEBUG
    {
      fprintf (stderr, "(DBG) lmp_eval_1: Got lmp_proc_error.\n");
#endif
      lmp_stop (lmp_last_token ());
#ifdef DEBUG
    }
#endif
  }

#ifdef DEBUG
  fprintf (stderr, "(DBG) lmp_eval: RETURN %Lf\n", left);
#endif
  return left;
}
